
# This file is automatically generated during the generation of setup.py
# Copyright 2020, Caer

author = 'Jason Dsouza: http://www.github.com/jasmcaus'
version = '1.5.1'
full_version = '1.5.1'
release = True
if not release:
    version = full_version



# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus

# Importing the necessary packages
import sys
sys.path.append('..')

import os
import time
import numpy as np
from .utils import readImg
from .utils import saveNumpy
from .preprocessing import MeanProcess

def preprocess_from_dir(DIR, classes, channels=1, IMG_SIZE=224, train_size=None, normalize_train=False, mean_subtraction=None, isShuffle=True, save_train=False, destination_filename=None, display_count=True):
    """
    Reads Images in base directory DIR using 'classes' 
    Returns
        train -> Image Pixel Values with corresponding labels (float32)
    Saves the above variables as .npy files if save_train = True
    """

    train = [] 
    if save_train:
        if destination_filename is None:
            raise ValueError('[ERROR] Specify a destination file name')

        elif not ('.npy' in destination_filename or '.npz' in destination_filename):
            raise TypeError('[ERROR] Specify the correct numpy destination file extension (.npy or .npz)', destination_filename)
    
        elif destination_filename is not None:
            destination_filename = None
    
    elif type(classes) is not list:
        raise ValueError('[ERROR] "classes" must be a list')

    elif not os.path.exists(DIR):
        raise ValueError('[ERROR] The specified directory does not exist', DIR)

    elif IMG_SIZE is None:
        raise ValueError('[ERROR] IMG_SIZE must be specified')

    elif type(IMG_SIZE) is not int:
        raise ValueError('[ERROR] IMG_SIZE must be an integer')

    # Loading from Numpy Files
    elif destination_filename is not None and os.path.exists(destination_filename):
        since = time.time()
        print('[INFO] Loading from Numpy Files')
        train = np.load(destination_filename, allow_pickle=True)
        end = time.time()
        print('----------------------------------------------')
        print('[INFO] Loaded in {:.0f}s from Numpy Files'.format(end-since))

        return train

    # Extracting image data
    else:
        since_preprocess = time.time()
        if destination_filename is not None:
            print(f'[INFO] Could not find {destination_filename}. Generating the Image Files')
        else:
            print('[INFO] Could not find a file to load from. Generating Image Files')
        print('----------------------------------------------')

        if train_size is None:
            train_size = len(os.listdir(os.path.join(DIR, classes[0])))

        # Checking if 'mean_subtraction' values are valid. Returns boolean value
        subtract_mean = check_mean_subtraction(mean_subtraction, channels)

        for item in classes:
            class_path = os.path.join(DIR, item)
            class_label = classes.index(item)
            count = 0 
            for image in os.listdir(class_path):
                if count != train_size:
                    image_path = os.path.join(class_path, image)

                    # Returns image RESIZED and img
                    img = readImg(image_path, IMG_SIZE=IMG_SIZE, channels=channels)
                    if img is None:
                        continue
                    # Normalizing
                    if normalize_train:
                        img = normalize(img)
                    
                    if subtract_mean:
                        mean_subtract = MeanProcess(mean_subtraction, channels)
                        img = mean_subtract.mean_preprocess(img, channels)
                        
                    train.append([img, class_label])
                    count +=1 

                    if display_count is True:
                        _printTotal(count, item)
                else:
                    break

        # Shuffling the Training Set
        if isShuffle is True:
            train = shuffle(train)

        # Converting to Numpy
        train = np.array(train)

        # Saves the Train set as a .npy file
        if save_train is True:
            #Converts to Numpy and saves
            if destination_filename.endswith('.npy'):
                print('[INFO] Saving as .npy file')
            elif destination_filename.endswith('.npz'):
                print('[INFO] Saving as .npz file')
            
            since = time.time()
            # Saving
            saveNumpy(destination_filename, train)
            end = time.time()
            
            time_elapsed = end-since

            print('[INFO] {} saved! Took {:.0f}m {:.0f}s'.format(destination_filename, time_elapsed // 60, time_elapsed % 60))

        #Returns Training Set
        end_preprocess = time.time()
        time_elapsed_preprocess = end_preprocess-since_preprocess
        print('----------------------------------------------')
        print('[INFO] Preprocessing complete! Took {:.0f}m {:.0f}s'.format(time_elapsed_preprocess // 60, time_elapsed_preprocess % 60))

        return train

def _printTotal(count, category):
    print(f'{count} - {category}')

def check_mean_subtraction(value, channels):
    """
        Checks if mean subtraction values are valid based on the number of channels
        Must be a tuple of dimensions = number of channels
    Returns boolean value
        True -> Expression is valid
        False -> Expression is invalid
    """
    if value is None:
        return False
    elif type(value) is tuple and len(value) == channels:
        return True
    else:
        raise ValueError(f'[ERROR] Expected a tuple of dimension {channels}', value) 

def shuffle(train):
    """
    Shuffles the Array
    """
    import random
    random.shuffle(train)
    return train

def sep_train(train, IMG_SIZE=None, channels=1):
    # x = []
    # y = []
    # for feature, label in train:
    #     x.append(feature)
    #     y.append(label)
    
    if IMG_SIZE is None:
        raise ValueError('[ERROR] IMG_SIZE not defined')
    else:
        x = [i[0] for i in train]
        y = [i[1] for i in train]

        # Without reshaping, X.shape --> (no. of images, IMG_SIZE, IMG_SIZE)
        # On reshaping, X.shape --> (no. of images, IMG_SIZE, IMG_SIZE,channels)

        # Converting to Numpy + Reshaping X
        x = reshape(x, IMG_SIZE, channels)
        y = np.array(y)

        return x, y

def reshape(x, IMG_SIZE, channels):
    return np.array(x).reshape(-1, IMG_SIZE, IMG_SIZE, channels)

def normalize(x, dtype='float32'):
    """
    Normalizes the data to mean 0 and standard deviation 1
    """
    # x/=255.0 raises a TypeError
    # x = x/255.0
    
    # Converting to float32 and normalizing (float32 saves memory)
    x = x.astype(dtype) / 255
    return x


# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus

# Importing the necessary packages
import cv2 as cv

class MeanProcess:
    def __init__(self, mean_sub_values, channels):
        # mean_sub_values is a tuple
        if channels == 3:
            self.rMean = mean_sub_values[0]
            self.gMean = mean_sub_values[1]
            self.bMean = mean_sub_values[2]
        if channels == 1:
            self.mean = mean_sub_values

    def mean_preprocess(self, image, channels):
        if channels == 3:
            (b,g,r) = cv.split(image.astype('float32')[:3])

            # Subtracting the mean
            r -= self.rMean
            b -= self.bMean
            g -= self.gMean

            # Merging 
            return cv.merge([b,g,r])
            
        if channels == 1:
            image -= self.mean
            return image


# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus

# Importing the necessary packages
from sklearn.feature_extraction.image import extract_patches_2d

class PatchPreprocess:
    def __init__(self, width, height):
        self.width = width
        self.height = height

    def patch_preprocess(self, image):
        return extract_patches_2d(image, (self.height, self.width), max_patches=1)[0]


# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus

from .version import version as __version__
from .version import author as __author__

from .preprocess import preprocess_from_dir
from .preprocess import sep_train
from .preprocess import shuffle
from .preprocess import reshape
from .preprocess import normalize
from .preprocess import _printTotal

from .io import HDF5Dataset
from .io import load_dataset

from .utils import readImg
from .utils import compute_mean
from .utils import saveNumpy
from .utils import train_val_split
from .utils import extract_frames
from .utils import plotAcc 
from .utils import url_to_image 
from .utils import toMatplotlib 
from .utils import translate
from .utils import rotate 
from .utils import rotate_bound 
from .utils import resize 
from .utils import canny 


# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus

#pylint: disable=bare-except

# Importing the necessary packages
# import os
from urllib.request import urlopen
import os
import time
import math
import cv2 as cv
import numpy as np

# # For Python 2.7
# import sys
# if sys.version_info.major == 2:
#     from urllib2 import urlopen
# For Python

def readImg(image_path, IMG_SIZE, channels=1):
    try:
        image_array = cv.imread(image_path)

        # [INFO] Using the following piece of code results in a 'None' in the training set
        # if image_array == None:
        #     pass
        if channels == 1:
            image_array = cv.cvtColor(image_array, cv.COLOR_BGR2GRAY)

        image_array = cv.resize(image_array, (IMG_SIZE,IMG_SIZE))
        return image_array
    except:
        return None

def compute_mean(DIR, channels):
    if channels == 3:
        rMean, gMean, bMean = 0,0,0
    if channels == 1:
        mean = 0
    count = 0

    if os.path.exists(DIR) is False:
        raise ValueError('The specified directory does not exist', DIR)

    for root, _, files in os.walk(DIR):
        for file in files:
            if file.endswith('.png') or file.endswith('.jpg'):
                count += 1
                filepath = root + os.sep + file
                img = cv.imread(filepath)
                if channels == 3:
                    b,g,r = cv.mean(img.astype('float32'))[:3]
                    rMean += r
                    bMean += b
                    gMean += g
                if channels == 1:
                    mean += cv.mean(img.astype('float32'))[0]

    # Computing average mean
    if channels == 3:
        rMean /= count
        bMean /= count 
        gMean /= count
        return rMean, bMean, gMean

    if channels == 1:
        mean /= count
        return mean

def saveNumpy(name, x):
    """
    Saves an array to a .npy file
    Converts to Numpy (if not already)
    """

    x = np.array(x)
    if '.npy' in name:
        np.save(name, x)
    elif '.npz' in name:
        np.savez_compressed(name, x)

def train_val_split(X,y,val_ratio=.2):
    """
    Returns X_train, X_val, y_train, y_val
    """
    try:
        from sklearn.model_selection import train_test_split    
        X_train, X_val, y_train, y_val = train_test_split(X,y,test_size=val_ratio,random_state = 2)
        return X_train, X_val, y_train, y_val
    except ModuleNotFoundError:
        print('[ERROR] The Sklearn Python package needs to be installed')

def extract_frames(input_folder, output_folder, IMG_SIZE=None, label_counter = None, video_count=None, frames_per_sec=None, dest_filetype='jpg'):
    """ Function to extract frames from videos within a directory
    and save them as separate frames in an output directory.
    Args:
        input_folder: Input video directory.
        output_folder: Output directory to save the frames.
        IMG_SIZE: Destination Image Size
        label_counter: Starting label counter
        video_count: Number of videos to process.
        frames_per_sec: Number of frames to process per second. 
        dest_filetype: Processed image filetype (png, jpg) --> Default: png
    Returns:
        label_counter (after processing)
    """

    dest_filetype.replace('.','')
    processed_videos = 0
    vid_count = 0 # to check if < video_count

    if os.path.exists(input_folder) is False:
        raise ValueError('Input folder does not exist', input_folder)

    if label_counter is None:
        label_counter = 0

    if video_count is None:
        video_count = len(os.listdir(input_folder))

    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    start = time.time()

    for root, _, files in os.walk(input_folder):
        for file in files:
            if vid_count < video_count:
                if file.endswith('.mp4') or file.endswith('.avi'): # if a video file
                    vid_filepath = root + os.sep + file
                    capture = cv.VideoCapture(vid_filepath)
                    video_frame_counter = 0

                    # Find the number of frames and FPS
                    video_frame_count = int(capture.get(cv.CAP_PROP_FRAME_COUNT)) - 1
                    video_fps = math.ceil(capture.get(cv.CAP_PROP_FPS))

                    print(f'{vid_count+1}. Reading \'{file}\'. Number of frames: {video_frame_count}. FPS: {video_fps}')

                    if frames_per_sec is not None:
                        interval= determine_interval(video_fps/frames_per_sec) # eg: 30//15
                        print('Interval: ', interval)
                    # if frames_per_sec is None, we assume that each frame should be processed
                    else:
                        interval = 1
                    
                    # Start converting the video
                    while capture.isOpened():
                        _, frame = capture.read()

                        if IMG_SIZE is not None:                    
                            frame = cv.resize(frame, (IMG_SIZE,IMG_SIZE))
                        
                        # Write the results back to output location as per specified frames per second
                        if video_frame_counter % interval == 0:
                            cv.imwrite(f'{output_folder}/{label_counter}.{dest_filetype}', frame)
                            video_frame_counter += 1
                            label_counter += 1
                            print('Frame counter: ', video_frame_counter)
                        
                        video_frame_counter += 1

                        # If there are no more frames left
                        if video_frame_counter > (video_frame_count-1):
                            capture.release()
                            processed_videos += 1
                            break

    end = time.time()
    # Printing stats
    taken = end-start
    print('[INFO] {} videos extracted in {:.2f} seconds'.format(processed_videos, taken ))

    return label_counter

def determine_interval(x):
    y = '{:.1f}'.format(x)
    inde = y.find('.') + 1
    if inde == -1: # if no '.' (if an integer)
        return x
    if int(y[inde]) < 5:
        return math.floor(x)
    else:
        return math.ceil(x)

def plotAcc(histories):
    """
    Plots the model accuracies as 2 graphs
    """
    import matplotlib.pyplot as plt 
    acc = histories.history['acc']
    val_acc = histories.history['val_acc']
    loss = histories.history['loss']
    val_loss = histories.history['val_loss']

    epochs = range(1, len(acc)+1)

    # Plotting Accuracy
    plt.plot(epochs, acc, 'b', label='Training Accuracy')
    plt.plot(epochs, val_acc, 'r', label='Validation Accuracy')
    plt.title('Training and Validation Accuracy')
    plt.legend()

    # Plotting Loss
    plt.plot(epochs, loss, 'b', label='Training Loss')
    plt.plot(epochs, val_loss, 'r', label='Validation Loss')
    plt.title('Training and Validation Loss')
    plt.legend()

    plt.show()

# -------------------- OPENCV IMAGE-SPECIFIC METHODS --------------------- 

def translate(image, x, y):
    # Defines the translation matrix and performs the translation
    matrix = np.float32([[1, 0, x], [0, 1, y]])
    shifted = cv.warpAffine(image, matrix, (image.shape[1], image.shape[0]))

    return shifted

def rotate(image, angle, centre=None, scale=1.0):
    # Grabs the dimensions of the image
    (height, width) = image.shape[:2]

    # If no centre is specified, we grab the centre coordinates of the image
    if centre is None:
        centre = (width // 2, height // 2)

    # Rotates the image
    M = cv.getRotationMatrix2D(centre, angle, scale)
    rotated = cv.warpAffine(image, M, (width, height))

    return rotated

def rotate_bound(image, angle):
    # Grabs the dimensions of the image and then determines the centre
    (h, w) = image.shape[:2]
    (cX, cY) = (w / 2, h / 2)

    # grab the rotation matrix (applying the negative of the
    # angle to rotate clockwise), then grab the sine and cosine
    # (i.e., the rotation components of the matrix)
    M = cv.getRotationMatrix2D((cX, cY), -angle, 1.0)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])

    # compute the new bounding dimensions of the image
    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))

    # Adjusts the rotation matrix to take into account translation
    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY

    # Performs the actual rotation and returns the image
    return cv.warpAffine(image, M, (nW, nH))

def resize(image, width=None, height=None, interpolation=cv.INTER_AREA):
    """
    Resizes the image while maintaing the aspect ratio of the original image
    """
    # Grabs the image dimensions 
    dim = None
    (h, w) = image.shape[:2]

    # if both the width and height are None, we return the original image
    if width is None and height is None:
        return image

    # If  width is None
    if width is None:
        # calculate the ratio of the height and construct the
        # dimensions
        r = height / float(h)
        dim = (int(w * r), height)

    # If height is None
    else:
        # Calculates the ratio of the width and constructs the dimensions
        r = width / float(w)
        dim = (width, int(h * r))

    # Resizes the image
    resized = cv.resize(image, dim, interpolation=interpolation)

    return resized

def toMatplotlib(image):
    """
    Converts BGR image ordering to RGB ordering
    """
    return cv.cvtColor(image, cv.COLOR_BGR2RGB)

def url_to_image(url, flag=cv.IMREAD_COLOR):
    # Converts the image to a Numpy array and reads it in OpenCV
    response = urlopen(url)
    image = np.asarray(bytearray(response.read()), dtype="uint8")
    image = cv.imdecode(image, flag)

    return image

def canny(image, sigma=0.33):
    # computes the median of the single channel pixel intensities
    med = np.median(image)

    # apply automatic Canny edge detection using the computed median
    lower = int(max(0, (1.0 - sigma) * med))
    upper = int(min(255, (1.0 + sigma) * med))
    edges = cv.Canny(image, lower, upper)

    return edges


# Author: Jason Dsouza
# Github: http://www.github.com/jasmcaus
# pylint: disable=unused-variable

import h5py 

class HDF5Dataset:
    def __init__(self, shape, dataset_name, buffer_size=1000):
        """
            shape holds the dimensions of the data. 
            For the MNIST Dataset, shape = (70000, 28, 28). This can be flattened into a feature vector (to be fed into any Machine Learning classifier like SVM or Logistic Regression) by reshaping to (70000, 784) --> 28*28 = 784
        """
        if '.h5' not in dataset_name:
            dataset_name += '.h5'
        
        # Creating 2 datasets -> Features/Images and corresponding labels
        self.db = h5py.File('dataset_name', 'w')

        self.features = self.db.create_dataset(name='features', shape=shape, dtype='float')
        self.labels = self.db.create_dataset(name='labels', shape=shape[0], dtype='int')

        # # We aren't adding data directly because that is handled via the buffer
        # self.features = self.db.create_dataset(name='features', shape=shape, data=X)
        # self.labels = self.db.create_dataset(name='labels', shape=shape[0], data=y)

        # Storing buffer size and initializing the buffer
        self.buffSize = buffer_size
        self.buffer = {'features': [], 'labels': []}
        self.buff_idx = 0
    
    def add(self, features, labels):
        # Adding rows and labels to the buffer
        self.buffer['features'].extend(features)
        self.buffer['labels'].extend(labels)

        # If buffer needs to be flushed to disk (if >buffer_size)
        if len(self.buffer['features']) >= self.buffSize:
            self.flush()
    
    def flush(self):
        # Buffers and written to disk and then flushed
        idx = self.buff_idx + len(self.buffer['features'])
        self.features[self.buff_idx : idx] = self.buffer['features']
        self.labels[self.buff_idx : idx] = self.buffer['labels']

        # Resetting buffer
        self.buff_idx = idx
        self.buffer = {'features': [], 'labels': []}
    
    def storeClassLabels(self, classLabels):
        """
            Stores the actual classLabels as an additional piece of the dataset
            classLabels must be a LIST
        """
        dtype = h5py.special_dtype(vlen='unicode')
        classLabels_set = self.db.create_dataset('class labels', shape=len(classLabels), dtype=dtype, data = classLabels)

    def close(self):
        # Writing any remaining items in buffer to disk
        if len(self.buffer['data']) > 0:
            self.flush()
        
        # Closing the dataset
        self.db.close()

# def create_dataset(X, y, dataset_name):
#     """
#     Creates an h5 dataset of features and corresponding labels
#     :param X: feature set
#     :param y: labels
#     :param dataset_name: Name for the dataset
#     """

#     if '.h5' not in dataset_name:
#         dataset_name += '.h5'

#     h5f = h5py.File('dataset_name', 'w')
#     h5f.create_dataset('features', data=X)
#     h5f.create_dataset('labels', data=y)
#     h5f.close()
    

def load_dataset(dataset_name):
    if '.h5' not in dataset_name:
        dataset_name = dataset_name + '.h5'
    
    h5f = h5py.File('dataset_name', 'r')
    X = h5f['features'][:] 
    y = h5f['labels'][:]
    h5f.close()

    return X, y


from setuptools import setup, find_packages
import sys
sys.path.append('..')

VERSION = '1.5.1'

DESCRIPTION = """A Computer Vision library in Python with powerful image processing operations, including support for Deep Learning models built using the Keras framework"""

LONG_DESCRIPTION = """A Computer Vision library in Python with powerful image processing operations, including support for Deep Learning models built using the Keras framework

This repository is actively being maintained. If there are any issues, kindly open a thread in the 'Issues' pane on the official Github repository. 
"""

AUTHOR = 'Jason Dsouza: http://www.github.com/jasmcaus'

VERSION_PY_TEXT = """
# This file is automatically generated during the generation of setup.py
# Copyright 2020, Caer

author = '%(author)s'
version = '%(version)s'
full_version = '%(full_version)s'
release = %(isrelease)s
if not release:
    version = full_version
"""

# Repository on PyPi.org = https://pypi.org/project/caer/


def write_version_py(filename='caer/version.py'):
    print('[INFO] Writing version.py')
    TEXT = VERSION_PY_TEXT
    FULL_VERSION = VERSION
    ISRELEASED = True

    a = open(filename, 'w')
    try:
        a.write(TEXT % {'author': AUTHOR,
                        'version': VERSION,
                       'full_version': FULL_VERSION,
                       'isrelease': str(ISRELEASED)})
    finally:
        a.close()


def setup_package():
    # Rewrite the version file everytime
    write_version_py()

    setup(
        name="caer",
        version=VERSION,
        author="Jason Dsouza",
        author_email="jasmcaus@gmail.com",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        url="https://github.com/jasmcaus/caer",
        download_url = "https://pypi.org/project/caer/",
        project_urls={
            "Bug Tracker": "https://github.com/jasmcaus/caer/issues",
            "Source Code": "https://github.com/jasmcaus/caer",
        },
        packages=find_packages(),
        license='MIT',
        install_requires=['numpy', 'opencv-contrib-python', 'h5py', 'sklearn'],
        keywords=['computer vision', 'deep learning', 'image processing', 'opencv', 'matplotlib'],
        classifiers=[
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Operating System :: MacOS",
            "Operating System :: Microsoft :: Windows",
            "License :: OSI Approved :: MIT License",
        ],
    )


if __name__ == '__main__':
    setup_package()